from builtins import str
import threading
import logging
import os

from biomaj.process.process import Process, DrmaaProcess, DockerProcess
from biomaj.mongo_connector import MongoConnector

class MetaProcess(threading.Thread):
    '''
    Meta process in biomaj process workflow. Meta processes are executed in parallel.

    Each meta process defined a list of Process to execute sequentially
    '''

    def __init__(self, bank, metas, meta_status=None, meta_data={}, simulate=False):
      '''
      Creates a meta process thread

      :param bank: Bank
      :type bank: :class:`biomak.bank`
      :param meta: list of meta processes to execute in thread
      :type meta: list of str
      :param meta_status: initial status of the meta processes
      :type meta_status: bool
      :param simulate: does not execute process
      :type simulate: bool
      '''
      threading.Thread.__init__(self)
      self._lock = None
      self.kill_received = False
      self.workflow = None
      self.simulate = simulate
      self.bank = bank
      self.metas = metas
      self.meta_data = meta_data
      self.meta_status = {}
      for meta in self.metas:
        self.meta_status[meta] = {}

      if meta_status is not None:
        self.meta_status = meta_status

      self._stopevent = threading.Event( )

      self.bmaj_env = os.environ.copy()
      #self.bmaj_env = {}
      # Copy all config from bank


      self.bmaj_only_env = {}
      #The root directory where all databases are stored.
      #If your data is not stored under one directory hirearchy
      #you can override this value in the database properties file.
      for conf in dict(self.bank.config.config_bank.items('GENERAL')):
        self.bmaj_env[conf] = self.bank.config.config_bank.get('GENERAL', conf)
        if self.bmaj_env[conf] is None:
          self.bmaj_env[conf] = ''
          self.bmaj_only_env[conf] = self.bmaj_env[conf]

      self.bmaj_env['dbname'] = self.bank.name
      self.bmaj_only_env['dbname'] = self.bmaj_env['dbname']

      self.bmaj_env['datadir'] = self.bank.config.get('data.dir')
      self.bmaj_only_env['datadir'] = self.bmaj_env['datadir']

      self.bmaj_env['data.dir'] = self.bmaj_env['datadir']
      self.bmaj_only_env['data.dir'] = self.bmaj_env['data.dir']

      if self.bank.config.get('mail.admin'):
          self.bmaj_env['mailadmin'] = self.bank.config.get('mail.admin')
          self.bmaj_only_env['mailadmin'] = self.bmaj_env['mailadmin']

      if self.bank.config.get('mail.smtp.host'):
        self.bmaj_env['mailsmtp'] = self.bank.config.get('mail.smtp.host')
        self.bmaj_only_env['mailsmtp'] = self.bmaj_env['mailsmtp']

      self.bmaj_env['processdir'] = self.bank.config.get('process.dir',default='')
      self.bmaj_only_env['processdir'] = self.bmaj_env['processdir']

      if 'PATH' in self.bmaj_env:
        self.bmaj_env['PATH'] += ':' + self.bmaj_env['processdir']
        self.bmaj_only_env['PATH'] = self.bmaj_env['PATH']
      else:
        self.bmaj_env['PATH'] = self.bmaj_env['processdir']+':/usr/local/bin:/usr/sbin:/usr/bin'
        self.bmaj_only_env['PATH'] = self.bmaj_env['PATH']

      self.bmaj_env['PP_DEPENDENCE'] = '#'
      self.bmaj_only_env['PP_DEPENDENCE'] = '#'
      self.bmaj_env['PP_DEPENDENCE_VOLATILE'] = '#'
      self.bmaj_only_env['PP_DEPENDENCE_VOLATILE'] = '#'
      self.bmaj_env['PP_WARNING'] = '#'
      self.bmaj_only_env['PP_WARNING'] = '#'


      self.bmaj_env['PATH_PROCESS_BIOMAJ'] = self.bank.config.get('process.dir')

      # Set some session specific env
      if self.bank.session is not None:
        self.bmaj_env['offlinedir'] = self.bank.session.get_offline_directory()
        self.bmaj_only_env['offlinedir'] = self.bmaj_env['offlinedir']

        self.bmaj_env['dirversion'] = self.bank.config.get('dir.version')
        self.bmaj_only_env['dirversion'] = self.bmaj_env['dirversion']

        self.bmaj_env['noextract'] = self.bank.config.get('no.extract')
        if self.bmaj_env['noextract'] is None:
          self.bmaj_env['noextract'] = ''
        self.bmaj_only_env['noextract'] = self.bmaj_env['noextract']

        self.bmaj_env['localrelease'] = self.bank.session.get_release_directory()
        self.bmaj_only_env['localrelease'] = self.bmaj_env['localrelease']
        if self.bank.session.get('release') is not None:
          self.bmaj_env['remoterelease'] = self.bank.session.get('remoterelease')
          self.bmaj_only_env['remoterelease'] = self.bmaj_env['remoterelease']
          self.bmaj_env['removedrelease'] = self.bank.session.get('release')
          self.bmaj_only_env['removedrelease'] = self.bmaj_env['removedrelease']

      for bdep in self.bank.depends:
        self.bmaj_env[bdep.name+'source'] = bdep.session.get_full_release_directory()
        self.bmaj_only_env[bdep.name+'source'] = self.bmaj_env[bdep.name+'source']


    def set_progress(self, name, status=None):
      '''
      Update progress on download

      :param name: name of process
      :type name: str
      :param status: status of process
      :type status: bool or None
      '''
      logging.debug('Process:progress:'+name+"="+str(status))
      if self.workflow is not None:
        MongoConnector.banks.update({'name': self.bank.name},{'$set': {'status.'+self.workflow+'.progress.'+name: status}})

    def run(self):
      # Run meta processes
      self.global_status = True
      for meta in self.metas:
        if not self._stopevent.isSet():
          logging.info("PROC:META:RUN:"+meta)
          processes = []
          if self.bank.config.get(meta) is not None:
            processes = self.bank.config.get(meta).split(',')
          processes_status = {}
          for bprocess in processes:
            if self.kill_received:
              raise Exception('Kill request received, exiting')
            # Process status already ok, do not replay
            if meta in self.meta_status and bprocess in self.meta_status[meta] and self.meta_status[meta][bprocess]:
              logging.info("PROC:META:SKIP:PROCESS:"+bprocess)
              processes_status[bprocess] = True
              continue
            logging.info("PROC:META:RUN:PROCESS:"+bprocess)
            # bprocess.name may not be unique
            #name = self.bank.config.get(bprocess+'.name')
            name = bprocess
            desc = self.bank.config.get(bprocess+'.desc')
            cluster = self.bank.config.get_bool(bprocess+'.cluster', default=False)
            docker = self.bank.config.get(bprocess+'.docker')
            proc_type = self.bank.config.get(bprocess+'.type')
            exe = self.bank.config.get(bprocess+'.exe')
            args = self.bank.config.get(bprocess+'.args')
            expand = self.bank.config.get_bool(bprocess+'.expand', default=True)
            if cluster:
              native = self.bank.config.get(bprocess+'.native')
              bmaj_process = DrmaaProcess(meta+'_'+name, exe, args, desc, proc_type, native, expand, self.bmaj_env, os.path.dirname(self.bank.config.log_file))
            elif docker:
              use_sudo = self.bank.config.get_bool('docker.sudo', default=True)
              bmaj_process = DockerProcess(meta+'_'+name, exe, args, desc, proc_type, docker, expand, self.bmaj_only_env, os.path.dirname(self.bank.config.log_file), use_sudo)
            else:
              bmaj_process = Process(meta+'_'+name, exe, args, desc, proc_type, expand, self.bmaj_env, os.path.dirname(self.bank.config.log_file))
            self.set_progress(bmaj_process.name, None)
            if self.bank.config.get(bprocess+'.format'):
              bmaj_process.format =  self.bank.config.get(bprocess+'.format')
            if self.bank.config.get(bprocess+'.types'):
              bmaj_process.types =  self.bank.config.get(bprocess+'.types')
            if self.bank.config.get(bprocess+'.tags'):
              bmaj_process.tags =  self.bank.config.get(bprocess+'.tags')
            if self.bank.config.get(bprocess+'.files'):
              bmaj_process.files =  self.bank.config.get(bprocess+'.files')
            res = bmaj_process.run(self.simulate)
            processes_status[bprocess] = res
            self.set_progress(bmaj_process.name, res)
            if not res:
              self.global_status = False
              break
            if not self.simulate:
              if self._lock:
                self._lock.acquire()
                try:
                  self._get_metata_from_outputfile(bmaj_process)
                except Exception as e:
                  logging.error(e)
                finally:
                  self._lock.release() # release lock, no matter what
              else:
                self._get_metata_from_outputfile(bmaj_process)
        self.meta_status[meta] = processes_status

    def _get_metata_from_outputfile(self, proc):
      '''
      Extract metadata given by process on stdout. Store metadata in self.metadata

      :param proc: process
      :type proc_name: :class:`biomaj.process.Process`
      '''
      proc_name = proc.name
      output_file = proc.output_file

      self.meta_data[proc_name] = {}
      with open(output_file) as f:
        for line in f:
          if line.startswith('##BIOMAJ#'):
            line = line.replace('##BIOMAJ#', '')
            line = line.strip('\n\r')
            metas = line.split('#')
            meta_format = metas[0]
            if meta_format == '':
              meta_format = proc.format
            meta_type = metas[1]
            if meta_type == '':
              meta_type = proc.types
            meta_tags = metas[2]
            if meta_tags == '':
              meta_tags = proc.tags
            meta_files = metas[3]
            if not meta_format in self.meta_data[proc_name]:
              self.meta_data[proc_name][meta_format] = []
            tags = meta_tags.split(',')
            tag_list = {}
            if meta_tags != '':
              for tag in tags:
                t = tag.split(':')
                tag_list[t[0]] = t[1]
            self.meta_data[proc_name][meta_format].append({'tags': tag_list,
                                                'types': meta_type.split(','),
                                                'files': meta_files.split(',')})
      if proc.files and proc.format:
        tag_list = {}
        if proc.tags != '':
          for tag in proc.tags.split(','):
              t = tag.split(':')
              tag_list[t[0]] = t[1]
        self.meta_data[proc_name][proc.format] = []
        self.meta_data[proc_name][proc.format].append({'tags': tag_list,
                                            'types': proc.types.split(','),
                                            'files': proc.files.split(',')})


    def stop(self):
      self._stopevent.set( )



from future import standard_library
standard_library.install_aliases()
import logging
import pycurl
import io
import re
import os

from biomaj.utils import Utils
from biomaj.download.ftp import FTPDownload

try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO

class HTTPDownload(FTPDownload):
  '''
  Base class to download files from HTTP

  Makes use of http.parse.dir.line etc.. regexps to extract page information

  protocol=http
  server=ftp.ncbi.nih.gov
  remote.dir=/blast/db/FASTA/

  remote.files=^alu.*\\.gz$

  '''

  def __init__(self, protocol, host, rootdir, config):
    FTPDownload.__init__(self, protocol, host, rootdir)
    self.config = config


  def list(self, directory=''):
    '''
    List FTP directory

    :return: tuple of file and dirs in current directory with details
    '''
    logging.debug('Download:List:'+self.url+self.rootdir+directory)
    self.crl.setopt(pycurl.URL, self.url+self.rootdir+directory)
    if self.credentials is not None:
      curl.setopt(pycurl.USERPWD, self.credentials)

    output = BytesIO()
    # lets assign this buffer to pycurl object
    self.crl.setopt(pycurl.WRITEFUNCTION, output.write)
    self.crl.setopt(pycurl.HEADERFUNCTION, self.header_function)
    self.crl.perform()
    # Figure out what encoding was sent with the response, if any.
    # Check against lowercased header name.
    encoding = None
    if 'content-type' in self.headers:
        content_type = self.headers['content-type'].lower()
        match = re.search('charset=(\S+)', content_type)
        if match:
             encoding = match.group(1)
    if encoding is None:
        # Default encoding for HTML is iso-8859-1.
        # Other content types may have different default encoding,
        # or in case of binary data, may have no encoding at all.
        encoding = 'iso-8859-1'

    # lets get the output in a string
    result = output.getvalue().decode(encoding)
    '''
    'http.parse.dir.line': r'<a[\s]+href="([\S]+)/".*alt="\[DIR\]">.*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})',
    'http.parse.file.line': r'<a[\s]+href="([\S]+)".*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})[\s]+([\d\.]+[MKG]{0,1})',
    'http.group.dir.name': 1,
    'http.group.dir.date': 2,
    'http.group.file.name': 1,
    'http.group.file.date': 2,
    'http.group.file.size': 3,
    '''

    rfiles = []
    rdirs = []

    dirs = re.findall(self.config.get('http.parse.dir.line'), result)
    if dirs is not None and len(dirs)>0:
      for dir in dirs:
        rfile = {}
        rfile['permissions'] = ''
        rfile['group'] = ''
        rfile['user'] = ''
        rfile['size'] = '0'
        date = dir[int(self.config.get('http.group.dir.date'))-1]
        dirdate = date.split()
        parts = dirdate[0].split('-')
        #19-Jul-2014 13:02
        rfile['month'] = Utils.month_to_num(parts[1])
        rfile['day'] = parts[0]
        rfile['year'] = parts[2]
        rfile['name'] = dir[int(self.config.get('http.group.dir.name'))-1]
        rdirs.append(rfile)

    files = re.findall(self.config.get('http.parse.file.line'), result)
    if files is not None and len(files)>0:
      for file in files:
        rfile = {}
        rfile['permissions'] = ''
        rfile['group'] = ''
        rfile['user'] = ''
        rfile['size'] = file[int(self.config.get('http.group.file.size'))-1]
        date = file[int(self.config.get('http.group.file.date'))-1]
        dirdate = date.split()
        parts = dirdate[0].split('-')
        #19-Jul-2014 13:02
        rfile['month'] = Utils.month_to_num(parts[1])
        rfile['day'] = parts[0]
        rfile['year'] = parts[2]
        rfile['name'] = file[int(self.config.get('http.group.file.name'))-1]
        rfiles.append(rfile)

    return (rfiles, rdirs)



from future import standard_library
standard_library.install_aliases()
from builtins import str
import logging
import pycurl
import io
import re
import os
from datetime import datetime

from biomaj.utils import Utils
from biomaj.download.interface import DownloadInterface

try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO


class FTPDownload(DownloadInterface):
  '''
  Base class to download files from FTP

  protocol=ftp
  server=ftp.ncbi.nih.gov
  remote.dir=/blast/db/FASTA/

  remote.files=^alu.*\\.gz$

  '''


  def __init__(self, protocol, host, rootdir):
    DownloadInterface.__init__(self)
    logging.debug('Download')
    self.crl = pycurl.Curl()
    url = protocol+'://'+host
    self.rootdir = rootdir
    self.url = url
    self.headers= {}


  def match(self, patterns, file_list, dir_list=[], prefix='', submatch=False):
    '''
    Find files matching patterns. Sets instance variable files_to_download.

    :param patterns: regexps to match
    :type patterns: list
    :param file_list: list of files to match
    :type file_list: list
    :param dir_list: sub directories in current dir
    :type dir_list: list
    :param prefix: directory prefix
    :type prefix: str
    :param submatch: first call to match, or called from match
    :type submatch: bool
    '''
    logging.debug('Download:File:RegExp:'+str(patterns))
    if not submatch:
        self.files_to_download = []
    for pattern in patterns:
      subdirs_pattern = pattern.split('/')
      if len(subdirs_pattern) > 1:
        # Pattern contains sub directories
        subdir = subdirs_pattern[0]
        if subdir == '^':
          subdirs_pattern = subdirs_pattern[1:]
          subdir = subdirs_pattern[0]
        for direlt in dir_list:
          subdir = direlt['name']
          logging.debug('Download:File:Subdir:Check:'+subdir)
          if pattern == '**/*':
            (subfile_list, subdirs_list) = self.list(prefix+'/'+subdir+'/')
            self.match([pattern], subfile_list, subdirs_list, prefix+'/'+subdir, True)
            for rfile in file_list:
              if pattern == '**/*' or re.match(pattern, rfile['name']):
                rfile['root'] = self.rootdir
                if prefix != '':
                  rfile['name'] = prefix + '/' +rfile['name']
                self.files_to_download.append(rfile)
                logging.debug('Download:File:MatchRegExp:'+rfile['name'])
          else:
            if re.match(subdirs_pattern[0], subdir):
              logging.debug('Download:File:Subdir:Match:'+subdir)
              # subdir match the beginning of the pattern
              # check match in subdir
              (subfile_list, subdirs_list) = self.list(prefix+'/'+subdir+'/')
              self.match(['/'.join(subdirs_pattern[1:])], subfile_list, subdirs_list, prefix+'/'+subdir, True)

      else:
        for rfile in file_list:
          if re.match(pattern, rfile['name']):
            rfile['root'] = self.rootdir
            if prefix != '':
              rfile['name'] = prefix + '/' +rfile['name']
            self.files_to_download.append(rfile)
            logging.debug('Download:File:MatchRegExp:'+rfile['name'])
    if not submatch and len(self.files_to_download) == 0:
      raise Exception('no file found matching expressions')

  def download(self, local_dir, keep_dirs=True):
    '''
    Download remote files to local_dir

    :param local_dir: Directory where files should be downloaded
    :type local_dir: str
    :param keep_dirs: keep file name directory structure or copy file in local_dir directly
    :param keep_dirs: bool
    :return: list of downloaded files
    '''
    logging.debug('FTP:Download')

    nb_files = len(self.files_to_download)
    cur_files = 1

    for rfile in self.files_to_download:
      if self.kill_received:
        raise Exception('Kill request received, exiting')
      file_dir = local_dir
      if 'save_as' not in rfile or rfile['save_as'] is None:
          rfile['save_as'] = rfile['name']
      if keep_dirs:
        file_dir = local_dir + '/' + os.path.dirname(rfile['save_as'])
      file_path = file_dir + '/' + os.path.basename(rfile['save_as'])

      self.mkdir_lock.acquire()
      try:
        if not os.path.exists(file_dir):
          os.makedirs(file_dir)
      except Exception as e:
        logging.error(e)
      finally:
        self.mkdir_lock.release() # release lock, no matter what
      logging.debug('FTP:Download:Progress:'+str(cur_files)+'/'+str(nb_files)+' downloading file '+rfile['name'])
      logging.debug('FTP:Download:Progress:'+str(cur_files)+'/'+str(nb_files)+' save as '+rfile['save_as'])
      cur_files += 1
      if not 'url' in rfile:
        rfile['url'] = self.url
      fp = open(file_path, "wb")
      curl = pycurl.Curl()
      curl.setopt(pycurl.URL, rfile['url']+rfile['root']+'/'+rfile['name'])
      if self.credentials is not None:
        curl.setopt(pycurl.USERPWD, self.credentials)
      curl.setopt(pycurl.WRITEDATA, fp)
      curl.perform()
      #errcode = curl.getinfo(pycurl.HTTP_CODE)
      #if int(errcode) != 200:
      #  self.error = True
      #  logging.error('Error while downloading '+rfile['name']+' - '+str(errcode))
      curl.close()
      fp.close()
      #logging.debug('downloaded!')
      self.set_permissions(file_path, rfile)
      # Add progress only per 10 files to limit db requests
      if nb_files < 10:
        nb = 1
        do_progress = True
      else:
        if cur_files == nb_files:
          do_progress = True
          nb = cur_files % 10
        elif cur_files > 0 and cur_files % 10 == 0:
          nb = 10
          do_progress= True
        else:
          do_progress = False
      if do_progress:
        self.set_progress(nb, nb_files)
    return self.files_to_download


  def header_function(self, header_line):
    # HTTP standard specifies that headers are encoded in iso-8859-1.
    # On Python 2, decoding step can be skipped.
    # On Python 3, decoding step is required.
    header_line = header_line.decode('iso-8859-1')

    # Header lines include the first status line (HTTP/1.x ...).
    # We are going to ignore all lines that don't have a colon in them.
    # This will botch headers that are split on multiple lines...
    if ':' not in header_line:
        return

    # Break the header line into header name and value.
    name, value = header_line.split(':', 1)

    # Remove whitespace that may be present.
    # Header lines include the trailing newline, and there may be whitespace
    # around the colon.
    name = name.strip()
    value = value.strip()

    # Header names are case insensitive.
    # Lowercase name here.
    name = name.lower()

    # Now we can actually record the header name and value.
    self.headers[name] = value


  def list(self, directory=''):
    '''
    List FTP directory

    :return: tuple of file and dirs in current directory with details
    '''
    logging.debug('Download:List:'+self.url+self.rootdir+directory)
    self.crl.setopt(pycurl.URL, self.url+self.rootdir+directory)
    if self.credentials is not None:
      curl.setopt(pycurl.USERPWD, self.credentials)
    output = BytesIO()
    # lets assign this buffer to pycurl object
    self.crl.setopt(pycurl.WRITEFUNCTION, output.write)
    self.crl.setopt(pycurl.HEADERFUNCTION, self.header_function)
    self.crl.perform()
    # Figure out what encoding was sent with the response, if any.
    # Check against lowercased header name.
    encoding = None
    if 'content-type' in self.headers:
        content_type = self.headers['content-type'].lower()
        match = re.search('charset=(\S+)', content_type)
        if match:
             encoding = match.group(1)
    if encoding is None:
        # Default encoding for HTML is iso-8859-1.
        # Other content types may have different default encoding,
        # or in case of binary data, may have no encoding at all.
        encoding = 'iso-8859-1'

    # lets get the output in a string
    result = output.getvalue().decode(encoding)
    # FTP LIST output is separated by \r\n
    # lets split the output in lines
    #lines = result.split(r'[\r\n]+')
    lines = re.split(r'[\n\r]+', result)
    # lets walk through each line
    rfiles = []
    rdirs = []
    for line in lines:
        rfile = {}
        # lets print each part separately
        parts = line.split()
        # the individual fields in this list of parts
        if not parts: continue
        rfile['permissions'] = parts[0]
        rfile['group'] = parts[2]
        rfile['user'] = parts[3]
        rfile['size'] = parts[4]
        rfile['month'] = Utils.month_to_num(parts[5])
        rfile['day'] = parts[6]
        try:
          rfile['year'] = int(parts[7])
        except Exception as e:
          # specific ftp case issues at getting date info
          curdate = datetime.now()
          rfile['year'] = curdate.year
          # Year not precised, month feater than current means previous year
          if rfile['month'] > curdate.month:
            rfile['year'] = curdate.year - 1
          # Same month but later day => previous year
          if rfile['month'] == curdate.month and int(rfile['day']) > curdate.day:
            rfile['year'] = curdate.year - 1
        rfile['name'] = parts[8]
        if len(parts) >= 10 and parts[9] == '->':
          # Symlink, add to files AND dirs as we don't know the type of the link
          rdirs.append(rfile)

        is_dir = False
        if re.match('^d', rfile['permissions']):
          is_dir = True

        if not is_dir:
          rfiles.append(rfile)
        else:
          rdirs.append(rfile)
    return (rfiles, rdirs)


  def chroot(self, cwd):
    logging.debug('Download: change dir '+cwd)

  def close(self):
    if self.crl is not None:
      self.crl.close()
      self.crl = None



from future import standard_library
standard_library.install_aliases()
from builtins import str
import datetime
import logging
import pycurl
import io
import os
import re
import urllib.request, urllib.parse, urllib.error

from biomaj.download.interface import DownloadInterface
from biomaj.download.ftp import FTPDownload
from biomaj.utils import Utils

try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO

class MultiDownload(DownloadInterface):
  '''
  Base interface for a downloader using multiple downloaders
  '''
  def __init__(self):
    DownloadInterface.__init__(self)
    self.downloaders = []
    self.files_to_download = []

  def add_downloaders(self, downloaders):
    '''
    Adds a list of downloaders
    '''
    self.downloaders += downloaders
    for d in downloaders:
      self.files_to_download += d.files_to_download


  def match(self, patterns, file_list, dir_list=[], prefix='', submatch=False):
    self.files_to_download = []
    for d in self.downloaders:
      d.match(patterns, d.files_to_download, [], prefix, submatch)
    self.files_to_download = []
    for d in self.downloaders:
      self.files_to_download += d.files_to_download

  def download(self, local_dir):
    self.files_to_download = []
    for d in self.downloaders:
      if self.kill_received:
        raise Exception('Kill request received, exiting')
      d.download(local_dir)
    self.files_to_download = []
    for d in self.downloaders:
      self.files_to_download += d.files_to_download
    return (self.files_to_download, [])

  def list(self):
    self.files_to_download = []
    for d in self.downloaders:
      d.list()
    self.files_to_download = []
    for d in self.downloaders:
      self.files_to_download += d.files_to_download
    return (self.files_to_download, [])

  def close(self):
    for d in self.downloaders:
      d.close()


class DirectFTPDownload(FTPDownload):
  '''
  download a list of files from FTP, no regexp
  '''

  def __init__(self, protocol, host, rootdir='', file_list=[]):
    '''

    Initialize the files in list with today as last-modification date.
    Size is also preset to zero, size will be set after download

    :param file_list: list of files to download on server
    :type file_list: list
    '''
    FTPDownload.__init__(self, protocol, host, rootdir)
    today = datetime.date.today()
    self.files_to_download = []
    self.headers = {}
    for file in file_list:
      rfile = {}
      rfile['root'] = self.rootdir
      rfile['permissions'] = ''
      rfile['group'] = ''
      rfile['user'] = ''
      rfile['size'] = 0
      rfile['month'] = today.month
      rfile['day'] = today.day
      rfile['year'] = today.year
      rfile['name'] = file
      self.files_to_download.append(rfile)

  def list(self, directory=''):
    '''
    FTP protocol does not give us the possibility to get file date from remote url
    '''
    return (self.files_to_download,[])

  def match(self, patterns, file_list, dir_list=[], prefix='', submatch=False):
    '''
    All files to download match, no pattern
    '''
    self.files_to_download = file_list



class DirectHttpDownload(DirectFTPDownload):

  def __init__(self, protocol, host, rootdir='', file_list=[]):
    '''
    :param file_list: list of files to download on server
    :type file_list: list
    '''
    DirectFTPDownload.__init__(self, protocol, host, rootdir, file_list)
    self.save_as = None
    self.method = 'GET'
    self.param = {}

  def download(self, local_dir, keep_dirs=True):
    '''
    Download remote files to local_dir

    :param local_dir: Directory where files should be downloaded
    :type local_dir: str
    :param keep_dirs: keep file name directory structure or copy file in local_dir directly
    :param keep_dirs: bool
    :return: list of downloaded files
    '''
    logging.debug('DirectHTTP:Download')
    nb_files = len(self.files_to_download)

    if nb_files > 1:
      self.files_to_download = []
      logging.error('DirectHTTP accepts only 1 file')

    cur_files = 1

    for rfile in self.files_to_download:
      if self.kill_received:
        raise Exception('Kill request received, exiting')

      if self.save_as is None:
        self.save_as = rfile['name']
      file_dir = local_dir
      if keep_dirs:
        file_dir = local_dir + os.path.dirname(self.save_as)
      file_path = file_dir + '/' + os.path.basename(self.save_as)
      self.mkdir_lock.acquire()
      try:
        if not os.path.exists(file_dir):
          os.makedirs(file_dir)
      except Exception as e:
        logging.error(e)
      finally:
        self.mkdir_lock.release() # release lock, no matter what
      logging.debug('DirectHTTP:Download:Progress'+str(cur_files)+'/'+str(nb_files)+' downloading file '+rfile['name']+', save as '+self.save_as)
      cur_files += 1
      if not 'url' in rfile:
        rfile['url'] = self.url
      fp = open(file_path, "wb")
      curl = pycurl.Curl()

      if self.method == 'POST':
        # Form data must be provided already urlencoded.
        postfields = urllib.parse.urlencode(self.param)
        # Sets request method to POST,
        # Content-Type header to application/x-www-form-urlencoded
        # and data to send in request body.
        if self.credentials is not None:
          curl.setopt(pycurl.USERPWD, self.credentials)

        curl.setopt(pycurl.POSTFIELDS, postfields)
        curl.setopt(pycurl.URL, rfile['url']+rfile['root']+'/'+rfile['name'])
      else:
        url = rfile['url']+rfile['root']+'/'+rfile['name']+'?'+urllib.parse.urlencode(self.param)
        curl.setopt(pycurl.URL, url)

      curl.setopt(pycurl.WRITEDATA, fp)
      curl.perform()

      curl.close()
      fp.close()
      logging.debug('downloaded!')
      rfile['name'] = self.save_as
      self.set_permissions(file_path, rfile)
      self.set_progress(1, nb_files)
    return self.files_to_download

  def header_function(self, header_line):
    # HTTP standard specifies that headers are encoded in iso-8859-1.
    # On Python 2, decoding step can be skipped.
    # On Python 3, decoding step is required.
    header_line = header_line.decode('iso-8859-1')

    # Header lines include the first status line (HTTP/1.x ...).
    # We are going to ignore all lines that don't have a colon in them.
    # This will botch headers that are split on multiple lines...
    if ':' not in header_line:
      return

    # Break the header line into header name and value.
    name, value = header_line.split(':', 1)

    # Remove whitespace that may be present.
    # Header lines include the trailing newline, and there may be whitespace
    # around the colon.
    name = name.strip()
    value = value.strip()

    # Header names are case insensitive.
    # Lowercase name here.
    name = name.lower()

    # Now we can actually record the header name and value.
    self.headers[name] = value

  def list(self, directory=''):
    '''
    Try to get file headers to get last_modification and size
    '''
    for file in self.files_to_download:
      self.crl.setopt(pycurl.HEADER, True)
      if self.credentials is not None:
        curl.setopt(pycurl.USERPWD, self.credentials)

      self.crl.setopt(pycurl.NOBODY, True)
      self.crl.setopt(pycurl.URL, self.url+self.rootdir+file['name'])
      output = BytesIO()
      # lets assign this buffer to pycurl object
      self.crl.setopt(pycurl.WRITEFUNCTION, output.write)
      self.crl.setopt(pycurl.HEADERFUNCTION, self.header_function)
      self.crl.perform()

      # Figure out what encoding was sent with the response, if any.
      # Check against lowercased header name.
      encoding = None
      if 'content-type' in self.headers:
          content_type = self.headers['content-type'].lower()
          match = re.search('charset=(\S+)', content_type)
          if match:
               encoding = match.group(1)
      if encoding is None:
          # Default encoding for HTML is iso-8859-1.
          # Other content types may have different default encoding,
          # or in case of binary data, may have no encoding at all.
          encoding = 'iso-8859-1'

      # lets get the output in a string
      result = output.getvalue().decode(encoding)

      lines = re.split(r'[\n\r]+', result)
      for line in lines:
        parts = line.split(':')
        if parts[0].strip() == 'Content-Length':
          file['size'] = parts[1].strip()
        if parts[0].strip() == 'Last-Modified':
          # Sun, 06 Nov 1994
          res = re.match('(\w+),\s+(\d+)\s+(\w+)\s+(\d+)', parts[1].strip())
          if res:
            file['day'] = res.group(2)
            file['month'] = Utils.month_to_num(res.group(3))
            file['year'] = res.group(4)
            continue
          #Sunday, 06-Nov-94
          res = re.match('(\w+),\s+(\d+)-(\w+)-(\d+)', parts[1].strip())
          if res:
            file['day'] = res.group(2)
            file['month'] = Utils.month_to_num(res.group(3))
            file['year'] = str(2000 + int(res.group(4)))
            continue
          #Sun Nov  6 08:49:37 1994
          res = re.match('(\w+)\s+(\w+)\s+(\d+)\s+\d{2}:\d{2}:\d{2}\s+(\d+)', parts[1].strip())
          if res:
            file['day'] = res.group(3)
            file['month'] = Utils.month_to_num(res.group(2))
            file['year'] = res.group(4)
            continue
    return (self.files_to_download,[])
